/*
 * (c) Copyright 2012 by Volker Bergmann. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted under the terms of the
 * GNU General Public License (GPL).
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * WITHOUT A WARRANTY OF ANY KIND. ALL EXPRESS OR IMPLIED CONDITIONS,
 * REPRESENTATIONS AND WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE
 * HEREBY EXCLUDED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.databene.platform.mongodb;

import org.databene.benerator.StorageSystem;
import org.databene.benerator.storage.AbstractStorageSystem;
import org.databene.commons.ConfigurationError;
import org.databene.commons.Context;
import org.databene.commons.StringUtil;
import org.databene.model.data.ComplexTypeDescriptor;
import org.databene.model.data.DataModel;
import org.databene.model.data.DefaultDescriptorProvider;
import org.databene.model.data.Entity;
import org.databene.model.data.TypeDescriptor;
import org.databene.webdecs.DataSource;
import org.databene.webdecs.util.ConvertingDataSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.mongodb.BasicDBObject;
import com.mongodb.DB;
import com.mongodb.DBCollection;
import com.mongodb.DBObject;
import com.mongodb.Mongo;

/**
 * {@link StorageSystem} implementation for Mongo DB.<br/><br/>
 * Created: 29.01.2012 07:56:13
 * @since 0.7.5
 * @author Volker Bergmann
 */
public class MongoDB extends AbstractStorageSystem {
	
	private static final Logger LOGGER = LoggerFactory.getLogger(MongoDB.class);
	
	private String id;
	private String host;
	private int port;
	private String database;
	private String user;
	private String password;
	
	private boolean connected;
	protected Mongo mongo;
	protected DB db;
	private Entity2DbObjectConverter entity2objectConverter;
	private DefaultDescriptorProvider descriptorProvider;
	private DbObject2IdConverter object2idConverter;
	
	// constructors ----------------------------------------------------------------------------------------------------
	
	public MongoDB(String id, DataModel dataModel) {
		setDataModel(dataModel);
		this.id = id;
		this.host = null;
		this.port = 27017;
		this.user = null;
		this.password = null;
		this.connected = false;
		this.mongo = null;
		this.db = null;
		this.entity2objectConverter = new Entity2DbObjectConverter();
		this.object2idConverter = new DbObject2IdConverter();
		this.descriptorProvider = new DefaultDescriptorProvider(id, dataModel);
	}

	// properties ------------------------------------------------------------------------------------------------------
	
	public String getId() {
		return id;
	}

	public void setId(String id) {
		this.id = id;
	}

	public String getHost() {
		return host;
	}

	public void setHost(String host) {
		this.host = host;
	}

	public int getPort() {
		return port;
	}

	public void setPort(int port) {
		this.port = port;
	}

	public String getDatabase() {
		return database;
	}

	public void setDatabase(String database) {
		this.database = database;
	}

	public String getUser() {
		return user;
	}

	public void setUser(String user) {
		this.user = user;
	}

	public String getPassword() {
		return password;
	}

	public void setPassword(String password) {
		this.password = password;
	}
	
	// DescriptorProvider interface implementation

	public TypeDescriptor[] getTypeDescriptors() {
		beConnected();
		return descriptorProvider.getTypeDescriptors();
	}

	public TypeDescriptor getTypeDescriptor(String typeName) {
		beConnected();
		return descriptorProvider.getTypeDescriptor(typeName);
	}
	
	// query methods from StorageSystem interface ----------------------------------------------------------------------

	public DataSource<Entity> queryEntities(String typeName, String selector, Context context) {
		LOGGER.debug("queryEntities({}, {})", typeName, selector);
		beConnected();
		BasicDBObject query = new BasicDBObject();
		if (!StringUtil.isEmpty(selector))
			throw new UnsupportedOperationException("'selector' is not yet supported in Mongo DB queries");
		// TODO define selector syntax (query.append("username", "johnd"));
		QueryDataSource source = new QueryDataSource(query, typeName, db);
		ComplexTypeDescriptor type = (ComplexTypeDescriptor) getTypeDescriptor(typeName);
		return new ConvertingDataSource<DBObject, Entity>(source, new DbObject2EntityConverter(this, type));
	}

	public DataSource<?> queryEntityIds(String type, String selector, Context context) {
		LOGGER.debug("queryEntityIds({}, {})", type, selector);
		beConnected();
		BasicDBObject query = new BasicDBObject();
		if (!StringUtil.isEmpty(selector))
			throw new UnsupportedOperationException("'selector' is not yet supported in Mongo DB queries");
		// TODO define selector syntax (query.append("username", "johnd"));
		QueryDataSource source = new QueryDataSource(query, type, db);
		return new ConvertingDataSource<DBObject, Object>(source, object2idConverter);
	}

	public DataSource<?> query(String selector, boolean simplify, Context context) {
		LOGGER.debug("query({})", selector);
		beConnected();
		throw new UnsupportedOperationException("General-purpose queries are not supported by Mongo DB");
	}
	
	// data manipulation methods from StorageSystem interface ----------------------------------------------------------
	
	@Override
	public Object execute(String command) {
		throw new UnsupportedOperationException("MongoSystem.execute() is not implemented"); // TODO implement MongoSystem.execute
	}
	
	public void store(Entity entity) {
		LOGGER.debug("store({})", entity);
		beConnected();
		String entityType = entity.type();
		if (entityType == null)
			throw new RuntimeException("Trying to persist an entity without type: " + entity);
		DBCollection collection = db.createCollection(entityType, null);
		DBObject doc = entity2objectConverter.convert(entity);
		collection.insert(doc);
	}

	public void update(Entity entity) {
		LOGGER.debug("update({})", entity);
		DBObject doc = entity2objectConverter.convert(entity);
		DBCollection collection = db.getCollection(entity.type());
		collection.save(doc);
	}
	
	// other methods from StorageSystem interface ----------------------------------------------------------------------
	
	public void flush() {
		// TODO does Mongo DB have any kind of bulk update?
	}
	
	public void close() {
		// TODO nothing to do here?
	}
	
	public void beConnected() {
		if (!connected) {
			try {
				this.mongo = new Mongo(host, port);
				this.db = mongo.getDB(database);
				if (!StringUtil.isEmpty(user) && !db.authenticate(user, StringUtil.getChars(password)))
					throw new ConfigurationError("Authentication failed for user '" + user + "' on " + host + ":" + port);
				for (String collectionName : db.getCollectionNames())
					descriptorProvider.addTypeDescriptor(new ComplexTypeDescriptor(collectionName, this));
				this.connected = true;
			} catch (Exception e) {
				throw new RuntimeException("Error connecting database '" + database + "' at " + host + ":" + port, e);
			}
		}
	}

	public ComplexTypeDescriptor getOrCreatePartType(String typeName) {
		ComplexTypeDescriptor type = (ComplexTypeDescriptor) getTypeDescriptor(typeName);
		if (type == null)
			type = new ComplexTypeDescriptor(typeName, this);
		return type;
	}
	
}
